#include <iostream>
#include <algorithm>
#include "board.h"

namespace
{
   struct find_room
   {
      std::string const *pCode;
      explicit find_room(std::string const &code) : pCode (&code) {}
      bool operator()(unit const &u) const {return (*pCode == u.place.code);}
   };

   struct compare_course
   {
      course const *pc;
      explicit compare_course(course const &co) : pc(&co) {}
      bool operator()(unit const &un) const {return (*pc == un.subject);}
   };

   struct find_clash
   {
      course const *pc;
      explicit find_clash(course const &co) : pc(&co) {}
      bool operator()(unit const &un) const
      {
         return (std::find(un.subject.clashgrp.begin(), un.subject.clashgrp.end(), 
                           pc->code) != un.subject.clashgrp.end());
      }
   };

   struct no_self_copy
   {
      std::string code;
      explicit no_self_copy(std::string const &c) : code(c) {}
      bool operator()(std::string const &c) const {return (code != c);}
   };
   
   struct degree_fit
   {
      int diff;
      unit un;

      explicit degree_fit(unit const &u)
         : un(u), diff(u.place.size - u.subject.size)
      {
         if (u.subject.owner != 0 && u.place.owner != u.subject.owner)
            diff += 1000;  // less preference given when course and room owner is not the same
      }

      bool operator<(const degree_fit &df) const {return (diff < df.diff);}
      bool operator()(unit const &u) const {return (un.place == u.place);}
      bool operator()(degree_fit const &df) const
      {
         return (un.place == df.un.place || un.subject == df.un.subject);
      }
   };

   bool SortByLocation(unit const &u1, unit const &u2)
   {
      return (u1.place.owner < u2.place.owner);
   }

   template <class In, class Out, class Pred>
   Out copy_if (In first, In last, Out res, Pred p)
   {
      while (first != last)
      {
         if (p(*first))
            *res++ = *first;

         ++first;
      }
      return res;
   } 
}

///////////////////////////////////////////////////////////////////////////////
// Clear all courses from the board
//    Arguments:
//       bLec - true to clear lecture units
//       bLab - true to clear lab units
///////////////////////////////////////////////////////////////////////////////
void board::ClearAllCourses(bool bLec, bool bLab)
{
   course reset;
   for (MAP_ITER posnet = net.begin(); posnet != net.end(); ++posnet)
   {
      for (UNITS_ITER posU = Units(posnet).begin(); posU != Units(posnet).end(); ++posU)
      {
         if (!(posU->bPermanent) && ((Place(posU).bLab && bLab) || (!Place(posU).bLab && bLec)))
            Subject(posU) = reset;
      }
   }
}

///////////////////////////////////////////////////////////////////////////////
// Determines if there is enough units for the number of credits to be fitted
//    Returns true if sufficient units, else return false
///////////////////////////////////////////////////////////////////////////////
bool board::IsSizeOk() const
{
   if (pts.size()==0 || net.size()==0)
      return false;

   int cRooms = 0;
   for (MAP_C_ITER posnet = net.begin(); posnet != net.end(); ++posnet)
      cRooms += static_cast<int>(posnet->second.units.size());

   int ccr = 0;
   for (COURSE_C_ITER pos = pts.begin(); pos != pts.end(); ++pos)
      ccr += pos->leccr;

   return (cRooms >= ccr);
}

///////////////////////////////////////////////////////////////////////////////
// Returns the total number of units in the board
///////////////////////////////////////////////////////////////////////////////
int board::BoardSize() const
{
   std::vector<unit>::size_type cSz = 0;
   for (MAP_C_ITER pos=net.begin(); pos != net.end(); ++pos)
      cSz += Units(pos).size();
   return static_cast<int>(cSz);
}

///////////////////////////////////////////////////////////////////////////////
// Randomly select a cell from the board
//    Arguments:
//       c    - course to fit
//       bLab - true if to fit a lab credit (needs 3 consecutive hours)
//    Returns iterators to the randomly selected event (dow and hour) and
//    cell (room)
///////////////////////////////////////////////////////////////////////////////
std::vector<std::pair<event, unit> > const board::CourseSchedule(course const &c) const
{
   std::vector<std::pair<event, unit> > vc;
   for (MAP_C_ITER posN = net.begin(); posN != net.end(); ++posN)
   {
      for (UNITS_C_ITER posU = Units(posN).begin(); posU != Units(posN).end(); ++posU)
      {
         if (Subject(posU) == c)
            vc.push_back(std::make_pair(Event(posN), *posU));
      }
   }
   return vc;
}

///////////////////////////////////////////////////////////////////////////////
// Randomly select a cell from the board
//    Arguments:
//       c    - course to fit
//       bLab - true if to fit a lab credit (needs 3 consecutive hours)
//    Returns iterators to the randomly selected event (dow and hour) and
//    cell (room)
///////////////////////////////////////////////////////////////////////////////
std::pair<board::MAP_ITER, board::UNITS_ITER> const board::RandomCell(course const &c, bool bLab)
{
   random rd;
   int cEvent = static_cast<int>(net.size());
   int nIdx = 0;
   MAP_ITER posnet;
   bool bOut = false;

   while (!bOut)
   {
      nIdx = (static_cast<int>(rd()) % cEvent);    // randomly pick an event
      posnet = net.begin();
      std::advance(posnet, nIdx);                  // move iterator to that chosen event
	
      if (bLab)
      {
         // fitting lab credits requires additional steps because a lab credit
         //    must be in 3 consecutive hours/units
         int dw = Dow(posnet);
         int hr = Hour(posnet);
         
         // work on a copy because we want to "see ahead" only
         MAP_ITER tempos = posnet;
         
         ++tempos;   // move to next hour
         if (tempos == net.end())
            continue;

         // 1 hour difference between previous hour and current hour
         //  and the chosen hour before the course hour limit?
         if (Dow(tempos) != dw || Hour(tempos) != (hr+1) || Hour(tempos) >= c.hourlimit)
            continue;

         ++tempos;   // move to next hour
         if (tempos == net.end() || Hour(tempos) >= c.hourlimit)
            continue;

         // 1 hour difference between previous hour and current hour?
         bOut = (Dow(tempos)==dw && Hour(tempos)==(hr+2));
      }
      else if (Hour(posnet) < c.hourlimit)
         bOut = true;
   }

   if (c.owner == 0)
      // owner == 0 can fit any room
      nIdx = (static_cast<int>(rd()) % rowSize_);
   else
   {
      // can only fit to rooms with owner == 0 or with same owner as course
      ROOMIDX_ loc;
      // randomly choose to fit to room with owner 0 or to room with same owner as course
      if (static_cast<int>(rd()) % 2 == 0) 
      {
         loc = roomidx_[0];         // fit to room with owner == 0
         if (loc.first == -1 || loc.last == -1) // room owner == 0 does not exist, so
            loc = roomidx_[c.owner];            //    fit to room with same owner as course
      }
      else
         loc = roomidx_[c.owner];   // fit to room with same owner as course

      // randomly choose the room to fit
      int x = static_cast<int>((rd() % 101));
      nIdx = static_cast<int>(x*(loc.last-loc.first)/100.0 + loc.first);
   }

   // move to the chosen room
   UNITS_ITER poscell = Units(posnet).begin() + nIdx;
   
   if (poscell->bPermanent)
      return RandomCell(c, bLab);   // unit cannot be changed; try random selection again
   
   return (std::make_pair(posnet, poscell));
}

///////////////////////////////////////////////////////////////////////////////
// Deletes a given course lab hours from the board
//    Arguments:
//       dow - day of week to delete the course lab hours
//       c   - course lab hours to delete (must pass by value not reference
//             because deletions will invalidate all pointers
///////////////////////////////////////////////////////////////////////////////
void board::DelLabUnits(int dow, course c)
{
   MAP_ITER pos = net.begin();
   while (pos != net.end() && Dow(pos) != dow)
      ++pos;   // move to the wanted dow

   while (pos != net.end() && Dow(pos) == dow)
   {
      // find the course
      UNITS_ITER posU = std::find_if(Units(pos).begin(), Units(pos).end(), compare_course(c));

      // delete only lab credits not lecture credits
      if (posU != Units(pos).end() && Place(posU).bLab)
         Units(pos).erase(posU);

      ++pos;   // next hour
   }
}

///////////////////////////////////////////////////////////////////////////////
// Adds a given course lab hours to the board
//    Arguments:
//       pos - iterators to three target hours (units) to fit course lab
//             IMPORTANT: these target hours must be checked for any
//             clashes before this member function is called
//       c   - course lab hours to add
//    Returns false if there already exists the same course lab hours in a
//    given day of week (dow); else true if successful addition
///////////////////////////////////////////////////////////////////////////////
bool board::AddLabUnits(MAP_ITER pos[3], course const &c)
{
   MAP_ITER posNet = net.begin();
   while (posNet != net.end() && Dow(posNet) != Dow(pos[0]))
      ++posNet;   // move to the wanted dow

   // search for any existing same course lab hours in the given dow
   while (posNet != net.end() && Dow(posNet) == Dow(pos[0]))
   {
      int hr = Hour(posNet);

      // only search in hours not equal to the 3 target hours;
      //    note: target hours must be checked for any clashes before this
      //    member function is called
      if (!(hr==Hour(pos[0]) || hr==Hour(pos[1]) || hr==Hour(pos[2])))
      {
         UNITS_ITER posU = std::find_if(Units(posNet).begin(), Units(posNet).end(), compare_course(c));
         // found existing course lab hours, so cannot continue
         if (posU != Units(posNet).end() && Place(posU).bLab)
            return false;
      }

      ++posNet;   // next hour
   }

   // no existing course lab hours in the given dow, so can add now:
   unit u;
   u.place.bLab = true;    // set to lab properties
   u.place.code = "LAB";
   u.place.name = "LAB";
   u.place.size = c.size;
   u.place.owner = c.owner;
   u.subject = c;
   Units(pos[0]).push_back(u);
   Units(pos[1]).push_back(u);
   Units(pos[2]).push_back(u);
   return true;
}

///////////////////////////////////////////////////////////////////////////////
// Preliminary check and setup of variables before fitting:
//    sets rowsize_ to the number of units in each row
//    sets roomidx_ to hold the positions of all room owners
///////////////////////////////////////////////////////////////////////////////
void board::PreFit()
{
   // ensure non-empty units and sufficient rooms for all courses
   if (!IsSizeOk())
      throw std::length_error("board::PreFit");

   // count the number of units for each row
   //    and make sure all rows have equal number of units
   MAP_ITER posN = net.begin();
   while (posN != net.end())
   {
      int cSizeT = static_cast<int>(Units(posN).size());
      if (posN == net.begin())
         rowSize_ = cSizeT;         // no. of units for the first row
      else if (rowSize_ != cSizeT)  // every row has equal no. of units?
         throw std::length_error("board::PreFit");

      ++posN;
   }

   ClearAllCourses();   // remove all courses; start fresh

   std::vector<int> own;
   for (posN = net.begin(); posN != net.end(); ++posN)
   {
      // sort so that rooms having the same owner are grouped
      std::sort(Units(posN).begin(), Units(posN).end(), SortByLocation);

      int start = 0, end = 0, nLoc = 0;   // owner position indexes
      bool bSet = true;
      std::map<int, ROOMIDX_> idx;
      std::vector<int> own2;

      // determine all room owner positions now
      for (UNITS_C_ITER i = Units(posN).begin(); i != Units(posN).end(); ++i)
      {
         if (bSet)
         {  
            // start of new room owner position
            start = end;
            nLoc = Place(i).owner;
            own2.push_back(nLoc);
            bSet = false;
         }
         else if (Place(i).owner != nLoc)
         {
            // room owner has changed so save previous owner position
            idx[nLoc] = ROOMIDX_(start, end);
            bSet = true;
         }

         // before exiting loop, add the last owner position
         if ((i+1) == Units(posN).end())
            idx[nLoc] = ROOMIDX_(start, end);
         else
            ++end;
      }

      if (posN == net.begin())
      {
         // first row, so set the variables; no checking yet
         roomidx_ = idx;
         own = own2;
      }
      else if (roomidx_ != idx || own2 != own)
         // all rows must have the same owners
         //    and the same owner positions
         throw std::length_error("board::PreFit");
   }

   // for each course, there must be a room with the same owner as that course
   //    room owner 0 fits all, so if 0 found, no need for further checking
   if (std::find(own.begin(), own.end(), 0) == own.end())
   {
      for (COURSE_C_ITER posC = pts.begin(); posC != pts.end(); ++posC)
      {
         if (std::find(own.begin(), own.end(), posC->owner) == own.end())
            throw std::domain_error("board::PreFit");
      }
   }
}

///////////////////////////////////////////////////////////////////////////////
// Fits all courses to board
//    Argument:
//    cAttempts - no. of fitting attempts (0 or 1 for a single atttempt only)
//
//    Returns true if fitting success or false if failure
///////////////////////////////////////////////////////////////////////////////
bool board::Fit(int cAttempts)
{
   nIter_ = 0;       // iteration counter for course fitting
   nLimit_ = rowSize_ * static_cast<int>(static_cast<double>(net.size()) * 0.1);
   if (nLimit_ < 1000)
      nLimit_ = 1000;  // make sure not too small value for max. no. of iterations

   std::map<event, row> origNet = net;    // save original condition

   random rd;
   int nAttempt = 1;
   bool bOk = false;
   while (!bOk)
   {
      bOk = true;    // assume fitting will be successful unless otherwise found
      int nCount = 0;
      std::random_shuffle(pts.begin(), pts.end(), rd); // random course selection

      // loop to fit each course
      for (COURSE_C_ITER i = pts.begin(); i != pts.end() && bOk; ++i)
      {
         if (fitOwner != -1 && fitOwner != i->owner)
            continue;

         std::cout << ++nCount << ". Now fitting course " << i->code << std::endl;
         int cLecCredits = i->leccr;
         int cLabCredits = i->labcr;
         while (bOk && (cLecCredits>0 || cLabCredits>0))
         {
            int bLec = static_cast<int>(rd()) % 2;    // fit lecture or lab?
            if (cLabCredits == 0)
               bLec = 1;   // fit lecture since no lab credits left
            else if (cLecCredits == 0)
               bLec = 0;   // fit lab since no lecture credits left
         
            // fit one credit at a time until no credits left to fit
            if (bLec == 0)
            {
               --cLabCredits;    // less one lab credit for fitting
               bOk = FitLab(*i);
            }
            else
            {
               --cLecCredits;    // less one lecture credit for fitting
               bOk = FitLecture(*i);
            }

            nIter_ = 0;    // reset iteration counter
         }
      }

      if (!bOk)
      {
         if (cAttempts > 0 && nAttempt < cAttempts)
         {
            net = origNet;       // restore to original condition before trying again
            std::srand(static_cast<unsigned int>(rd()));    // new fitting attempt
            ++nAttempt;
            std::cout << nAttempt << ". *** FAILED fitting -- trying another set of combinations ***" << std::endl;
         }
         else
            break;
      }
   }

   if (bOk)   
      PostFit();

   return bOk;
}

///////////////////////////////////////////////////////////////////////////////
// Post fitting to ensure better course-room fittings/combinations
///////////////////////////////////////////////////////////////////////////////
void board::PostFit()
{
   // loop to find the best room to fit each course
   //    criteria:
   //    = course should be fitted to a room having the same course owner
   //    = find the smallest difference between room size and course size
   for (MAP_ITER posN = net.begin(); posN != net.end(); ++posN)
   {
      std::vector<degree_fit> df;   // holds the size differences
      std::vector<room> vr;
      std::vector<course> vc;

      // fill containers with all courses and rooms for current row
      for (UNITS_ITER posU = Units(posN).begin(); posU != Units(posN).end(); ++posU)
      {
         // skip labs and permanent units
         if (!(Place(posU).bLab) && !(posU->bPermanent))
         {
            vr.push_back(Place(posU));
            if (!(Subject(posU).code.empty()))     // skip empty units
               vc.push_back(Subject(posU));
         }
      }

      if (vc.empty())   // next row if current row has no courses
         continue;

      // calculate the size differences for all room-course combinations
      for (COURSE_ITER posC = vc.begin(); posC != vc.end(); ++posC)
      {
         for (std::vector<room>::iterator posR = vr.begin();
              posR != vr.end(); ++posR)
         {
            // calculate size difference only if room can fit course,
            //    owner is 0, or if course owner is the same as room owner
            if (posR->size >= posC->size && 
                (posC->owner == 0 || posR->owner == 0 || posR->owner == posC->owner))
            {
               unit u;
               u.place = *posR;
               u.subject = *posC;
               df.push_back(degree_fit(u));
            }
         }
      }

      // sort from the smallest to largest size difference
      std::sort(df.begin(), df.end());
      std::vector<degree_fit>::iterator posDU = df.begin();
      int nIdx = 0;

      // once the smallest size difference for a given course-room combination is
      //    chosen, remove all other combinations involving that room or course
      while (posDU != df.end())
      {
         degree_fit comp(*posDU);   // smallest difference first
         ++posDU;                   // move to next room-course combination
         if (posDU != df.end())
         {
            // remove duplicate rooms and courses
            df.erase(std::remove_if(posDU, df.end(), comp), df.end());
            posDU = df.begin();  // reset to the begining
            ++nIdx;
            posDU += nIdx;       // move to next element
         }
      }

      // clear all lecture courses to fit new room-course combination
      Row(posN).ClearCourses(true, false);
      for (posDU = df.begin(); posDU != df.end(); ++posDU)
      {
         // find room
         UNITS_ITER ui = std::find_if(Units(posN).begin(), Units(posN).end(), degree_fit(*posDU));
         if (ui != Units(posN).end())
            Subject(ui) = posDU->un.subject;    // room found, now fit course
      }
   }
}

///////////////////////////////////////////////////////////////////////////////
// For a given course, counts the number of lecture and lab credits in a given
//    day of week (dow)
//    Arguments:
//    oc - the number of credits found is returned
//    e  - day of week (dow) to start counting; hour is not used
//    c  - the course to find
//    
//    Returns true if a course lecture only occurs once (or none), 
//    course lab hours are on three consecutive hour stretch (or none), and
//    all courses placed in the correct rooms (i.e., no mismatch owners);
//    else returns false is the above condition is not met
///////////////////////////////////////////////////////////////////////////////
bool board::Count(OCCURRED_ &oc, const event &e, course const &c)
{
   oc.numLec = 0;    // reset
   oc.numLab = 0;

   MAP_C_ITER posNet = net.begin();
   while (posNet != net.end() && Dow(posNet) != e.dow)
      ++posNet;      // move iterator until the wanted dow is found

   std::vector<int> labhours;    // holds lab hours to check if lab hours are consecutive
   // count credits for all hours in a given dow
   while (posNet != net.end() && Dow(posNet) == e.dow)
   {
      // count credits in a given hour
      for (UNITS_C_ITER posU = Units(posNet).begin();
           posU != Units(posNet).end(); ++posU)
      {
         if (Subject(posU) == c)    // course found?
         {
            // valid if owner is 0 or course and room have the same owner
            bool bAlloc = (c.owner == 0 || Place(posU).owner == 0 || Place(posU).owner == c.owner);
            if (!bAlloc)
            {
               return false;
            }

            if (Place(posU).bLab)   // lab or lecture?
            {
               ++oc.numLab;
               labhours.push_back(Hour(posNet));   // check later if lab hours are consecutive
            }
            else
               ++oc.numLec;
         }
      }

      ++posNet;   // next hour
   }

   // lecture must only occur once (or none) in a dow, and lab hour
   //    count must be 3 (or 0) hours only (occurring once or none in a dow)
   bool bValid = ((oc.numLab==0 || oc.numLab==3) && oc.numLec <= 1);
   if (bValid)
   {
      oc.numLab = static_cast<int>(oc.numLab / 3);    // will be either 0 or 1
      if (oc.numLab == 1)
      {
         // are lab hours consecutive?
         bValid = (labhours[1]-labhours[0] == 1 && labhours[2]-labhours[1] == 1);
      }
   }

   return bValid;
}

///////////////////////////////////////////////////////////////////////////////
// Checks validity of the board
//    Argument:
//    bIgnore - set to true to fully check any course clashes; else set to
//              false to ignore clashes between permanent courses
//
//    Returns true if valid; false if invalid
///////////////////////////////////////////////////////////////////////////////
bool board::IsValid(bool bIgnore)
{
   bool bValid = IsSizeOk();  // empty units?

   // 1. any clashes for a given row?
   MAP_C_ITER posNet = net.begin();
   for (; posNet != net.end() & bValid; ++posNet)
      bValid = Row(posNet).IsValid(bIgnore);

   // 2. all credits (lecture and lab) for a given course must be accounted for
   for (COURSE_C_ITER posP = pts.begin(); posP != pts.end() && bValid; ++posP)
   {
      if (fitOwner != -1 && fitOwner != posP->owner)
         continue;

      OCCURRED_ oc;      // will hold the counted lecture and lab credits
      posNet = net.begin();
      while (posNet != net.end() && bValid)  // check all hours for each dow
      {
         event e = Event(posNet);
         OCCURRED_ temp;
         bValid = Count(temp, e, *posP);     // count credits for a given course
         
         oc.numLec += temp.numLec;     // add up the credits counted previously
         oc.numLab += temp.numLab;
         while (bValid && posNet != net.end() && Dow(posNet) == e.dow)
            ++posNet;   // move to next dow
      }

      if (bValid)
      {
         // all credits accounted for?
         bValid = (oc.numLec == posP->leccr && oc.numLab == posP->labcr);
      }
   }

   return bValid;
}

///////////////////////////////////////////////////////////////////////////////
// Fits one (any only one) lab credit (3 hours). Call this function repeatedly
//    to fit any remaining lab credits
//    Argument:
//    c - course to fit
//    
//    Returns true if successful fitting; else returns false
///////////////////////////////////////////////////////////////////////////////
bool board::FitLab(course const &c)
{
   while (++nIter_ <= nLimit_)
   {
      // randomly choose an event and room
      std::pair<MAP_ITER, UNITS_ITER> newpos = RandomCell(c, true);

      // holds the three hours for the one lab credit:
      MAP_ITER poshr[3];
      poshr[0] = newpos.first;
      poshr[1] = poshr[0];
      ++poshr[1];
      poshr[2] = poshr[1];
      ++poshr[2];

      // course clashes for the three hours:
      std::vector<std::string> clash[3];
      clash[0] = Row(poshr[0]).GroupClashes();
      clash[1] = Row(poshr[1]).GroupClashes();
      clash[2] = Row(poshr[2]).GroupClashes();

      // count the number of clash for each of those 3 hours:
      int numClash[3];
      numClash[0] = static_cast<int>(std::count(clash[0].begin(), clash[0].end(), c.code));
      numClash[1] = static_cast<int>(std::count(clash[1].begin(), clash[1].end(), c.code));
      numClash[2] = static_cast<int>(std::count(clash[2].begin(), clash[2].end(), c.code));

      // try another hour if too many clashes; must be none
      //    or one clash only for each hour -- don't disturb the board too much
      //    by displacing too many courses just to fit the target course
      if (numClash[0]>1 || numClash[1]>1 || numClash[2]>1)
      {
         continue;
      }

      // Possibilities:
      //    1) No clash for all the 3 hours
      //    2) 0 or 1 clash for one or more of the 3 hours
      if (numClash[0]==0 && numClash[1]==0 && numClash[2]==0)
      {
         // Possibility (1): No clash for all the 3 hours
         if (AddLabUnits(poshr, c))
         {
            return true;   // successful fitting; else try another set of 3 hours
         }
      }
      else
      {
         // Possibility (2): 0 or 1 clash for one or more of the 3 hours
         int i = 0;
         while (numClash[i] == 0)
            ++i;     // find the hour that has the first clash

         // find out what is causing the clash for that hour:
         // Possibilities:
         //    a) Clash is due to same course (a lecture), or
         //    b) Clash is due to another course (AC):
         //       i)  AC is a lab
         //       ii) AC is a lecture
         UNITS_ITER posU = std::find_if(Units(poshr[i]).begin(), Units(poshr[i]).end(), compare_course(c));
         
         if (posU != Units(poshr[i]).end() && !Place(posU).bLab)
         {
            // Possibility (2a): clash is due to the same course (a lecture),
            //    so must ensure other hours are clash-free; must only displace
            //    one course -- minimal board disturbance
            int idx = i + 1;
            bool bOtherZeroes = true;
            while (idx < 3 && bOtherZeroes)
            {
               bOtherZeroes = (numClash[idx] == 0);  // subsequent hours clash-free?
               ++idx;
            }

            // re-fit the displaced course if all other hours are clash-free;
            //    else try another set of 3 hours
            if (bOtherZeroes && AddLabUnits(poshr, c))
            {
               // find the clashed course (lecture), displaced it, and re-fit it elsewhere:
               posU = std::find_if(Units(poshr[i]).begin(), Units(poshr[i]).end(), compare_course(c));
               Subject(posU) = course();     // reset

               return FitLecture(c);   // re-fit the course lecture
            }
         }
         else if (posU == Units(poshr[i]).end())
         {
            // Possibility (2b): clash is due to another course (AC)
            posU = std::find_if(Units(poshr[i]).begin(), Units(poshr[i]).end(), find_clash(c)); // find AC
            
            // cannot displace AC if, compared to the target course,
            //    1) AC's hour limit is smaller (more restrictive fitting for AC), or
            //    2) AC has more clashes (provided it has equal hour limits with
            //       the target course); AC has more clashes, so more risks of clashes
            //       if re-fit elsewhere
            if (Subject(posU).clashgrp.size() > c.clashgrp.size() ||
                Subject(posU).hourlimit < c.hourlimit ||
                (Subject(posU).hourlimit == c.hourlimit && 
                 Subject(posU).clashgrp.size() > c.clashgrp.size()))
               continue;
            
            // is AC a lab or lecture?
            if (Place(posU).bLab)
            {
               // Possibility (2b-i): AC is a lab
               //    can displace AC only if the clashes for the subsequent hours are due
               //    to the same course AC (lab). Try another set of 3 hours if other clashes
               //    are by courses other than AC or by the same course as AC but the
               //    clashed course is a lecture (not a lab) -- minimal board disturbance.
               bool bOk = true;
               for (int idx = i + 1; bOk && idx < 3; ++idx)
               {
                  if (numClash[idx] != 0)
                  {
                     UNITS_C_ITER posOther = std::find_if(Units(poshr[idx]).begin(),
                                                          Units(poshr[idx]).end(), 
                                                          compare_course(Subject(posU)));
                     // same course as AC and a lab?
                     bOk = (posOther != Units(poshr[idx]).end() && Place(posOther).bLab);
                  }
               }

               if (bOk && AddLabUnits(poshr, c))
               {
                  // find AC and remove all its lab hours
                  posU = std::find_if(Units(poshr[i]).begin(), Units(poshr[i]).end(), find_clash(c));
                  course displaced = Subject(posU);
                  DelLabUnits(Dow(poshr[i]), Subject(posU));

                  return FitLab(displaced);  // re-fit the displaced AC lab hours
               }
            }
            else
            {
               // Possibility (2b-ii): AC is a lecture,
               //    so must ensure other hours are clash-free; must only displace
               //    one course -- minimal board disturbance
               int idx = i + 1;
               bool bOtherZeroes = true;
               while (idx < 3 && bOtherZeroes)
               {
                  bOtherZeroes = (numClash[idx] == 0);  // subsequent hours clash-free?
                  ++idx;
               }

               // re-fit AC if all other hours are clash-free;
               //    else try another set of 3 hours
               if (bOtherZeroes && AddLabUnits(poshr, c))
               {
                  // find the AC (lecture), displaced it, and re-fit it elsewhere:
                  posU = std::find_if(Units(poshr[i]).begin(), Units(poshr[i]).end(), find_clash(c));
                  course displaced = Subject(posU);
                  Subject(posU) = course();     // reset
   
                  return FitLecture(displaced);    // re-fit the displaced AC lecture hour
               }
            }
         }
      }
   }

   return false;  // max. iteration reached and still failed to fit
}

///////////////////////////////////////////////////////////////////////////////
// Fits one (any only one) lecture credit. Call this function repeatedly to fit
//    any remaining lecture credits
//    Argument:
//    c - course to fit
//    
//    Returns true if successful fitting; else returns false
///////////////////////////////////////////////////////////////////////////////
bool board::FitLecture(course const &c)
{
   while (++nIter_ <= nLimit_)
   {
      // randomly choose an event and room
      std::pair<MAP_ITER, UNITS_ITER> newpos = RandomCell(c);

      // choose again if room is a lab, room cannot fit the course,
      //    or owner mismatch (except owner == 0)
      if (Place(newpos.second).bLab || Place(newpos.second).size < c.size ||
          (c.owner != 0 && Place(newpos.second).owner != 0 && 
           Place(newpos.second).owner != c.owner))
         continue;

      // choose again if any clashes
      std::vector<std::string> clash = Row(newpos.first).GroupClashes();
      if (std::find(clash.begin(), clash.end(), c.code) != clash.end())
         continue;

      // Possibilities:
      //    1) target room is unoccupied, or
      //    2) room is occupied by another course
      if (Subject(newpos.second).code.empty())
      {
         // Possibility (1): chosen target room is unoccupied
         MAP_ITER pos = net.begin();
         while (pos != net.end() && Dow(pos) != Dow(newpos.first))
            ++pos;   // move to the given dow

         // before adding the target course, make sure the same course doesn't exist
         //    anywhere in the given dow, unless the existing course doesn't fit the
         //    room as well as the target course fits the target room
         bool bAdd = true;
         while (pos != net.end() && Dow(pos) == Dow(newpos.first) && bAdd)
         {
            // find any existing same course in all hours of the given dow
            //    except for target hour since it was already checked for 
            //    clash earlier
            if (Hour(pos) != Hour(newpos.first))
            {
               UNITS_ITER loc = std::find(Units(pos).begin(), Units(pos).end(), c);
               if (loc != Units(pos).end())
               {
                  // found an existing same course elsewhere,
                  //    so now check degree of fit
                  int iOldDiff = Place(loc).size - c.size;
                  int iNewDiff = Place(newpos.second).size - c.size; 
                  if (iNewDiff < iOldDiff)
                  {
                     // target fitting is better than the fitting of the
                     //    existing course, so displace existing course
                     //    and fit the target course to the target room
                     Subject(newpos.second) = c;
                     Subject(loc) = course();      // reset

                     return FitLecture(c);   // re-fit the displaced course
                  }
                  else
                     bAdd = false;     // target fitting is worse than the
                                       //    fitting of the existing course, so
                                       //    exit inner loop to try another room
               }
            }

            ++pos;
         }

         if (bAdd)
         {
            // no existing same course found, so just add target course
            //    to the unoccupied room
            Subject(newpos.second) = c;
            return true;      // successful fitting
         }
      }
      else
      {
         // Possibility (2): chosen target room is occuppied by another course (AC)

         // can displace AC if, as compared to target course,
         //   1) no. of clashes for AC is smaller (AC has less clashes, so lesser risks
         //      of clashes if re-fit elsewhere),
         //   2) hour limit of AC is larger (fitting AC to any hour is less restrictive), or
         //   3) degree of room fit for AC is larger (room is too large for AC)
         if (c.clashgrp.size() > Subject(newpos.second).clashgrp.size() ||
//             c.hourlimit < Subject(newpos.second).hourlimit ||
             ((Place(newpos.second).size - Subject(newpos.second).size) > 
              (Place(newpos.second).size - c.size)))
         {
            bool bFound = false;
            for (MAP_ITER pos = net.begin(); pos != net.end() && !bFound; ++pos)
            {
               // find any existing same course in all hours of the given dow
               //    except for target hour since it was already checked for any
               //    clash earlier. Need to search because we don't want more than one target
               //    course (> 1 lecture credit) in a dow.
               if (Dow(pos) == Dow(newpos.first) && Hour(pos) != Hour(newpos.first))
               {
                  UNITS_ITER loc = std::find(Units(pos).begin(), Units(pos).end(), c);
                  bFound = (loc != Units(pos).end());
               }
            }

            if (!bFound)
            {
               // there isn't any existing same target course anywhere in the given dow,
               //    so can displace AC
               course displaced = Subject(newpos.second);
               Subject(newpos.second) = c;
               
               return FitLecture(displaced);    // re-fit the displaced AC
            }
         }
      }
   }

   return false;  // max. iteration reached and still failed to fit
}

void board::ReadCourses(std::ifstream &in)
{
   // save current state of exception flags; this function can throw exceptions
   std::ios_base::iostate oldEx = in.exceptions();
   in.exceptions(std::ios_base::failbit|std::ios_base::badbit);

   pts.clear();
   int cCourse;
   in >> cCourse;
   for (int i = 0; i < cCourse; ++i)
   {
      course sub;
      in >> sub.code;
      in >> sub.name;
      in >> sub.size;
      in >> sub.owner;
      in >> sub.leccr;
      in >> sub.labcr;
      in >> sub.hourlimit;
      pts.push_back(sub);
   }

   in.exceptions(oldEx);
}

void board::ReadBoard(BOARDFILES &bf)
{
   // save current state of exception flags; this function can throw exceptions
   std::ios_base::iostate oldEx = bf.fEvent.exceptions();
   bf.fEvent.exceptions(std::ios_base::failbit | std::ios_base::badbit | std::ios_base::eofbit);
   int cDow;
   bf.fEvent >> cDow;
   std::vector<int> vdow;
   int i = 0;
   for (; i < cDow; ++i)
   {
      int dow;
      bf.fEvent >> dow;
      vdow.push_back(dow);
   }

   int cHour;
   bf.fEvent >> cHour;
   std::vector<int> vhour;
   for (i = 0; i < cHour; ++i)
   {
      int hour;
      bf.fEvent >> hour;
      vhour.push_back(hour);
   }

   bf.fEvent.exceptions(oldEx);
   oldEx = bf.fRoom.exceptions();
   bf.fRoom.exceptions(std::ios_base::failbit | std::ios_base::badbit | std::ios_base::eofbit);
   int cRoom;
   bf.fRoom >> cRoom;
   row r;
   for (i = 0; i < cRoom; ++i)
   {
      unit u;
      bf.fRoom >> u.place.code;
      bf.fRoom >> u.place.name;
      bf.fRoom >> u.place.size;
      bf.fRoom >> u.place.owner;
      r.units.push_back(u);
   }

   bf.fRoom.exceptions(oldEx);

   net.clear();
   for (i = 0; i < cDow; ++i)
   {
      for (int j = 0; j < cHour; ++j)
          net[event(vdow[i], vhour[j])] = r;
   }

   ReadCourses(bf.fCourse);
   if (bf.fClash)
      ReadClashes(bf.fClash);

   PreFit();

   if (bf.fPermanent)
      ReadPermanents(bf.fPermanent);
}

void board::ReadPermanents(std::ifstream &in)
{
   // save current state of exception flags; this function can throw exceptions
   std::ios_base::iostate oldEx = in.exceptions();
   in.exceptions(std::ios_base::failbit | std::ios_base::badbit | std::ios_base::eofbit);

   int cPermanent;
   in >> cPermanent;                      // no. of permanent courses to read
   for (int i = 0; i < cPermanent; ++i)   // loop to read each permanent course
   {
      std::string code;
      in >> code;             // first, read course code string
      course tcourse;
      tcourse.code = code;    // next, find the course object
      COURSE_ITER pos = std::find(pts.begin(), pts.end(), tcourse);
      // permanent course must be found (already called ReadCourses()?)
      if (pos == pts.end())
         throw std::runtime_error("board::ReadPermanents");

      int cEvents;
      in >> cEvents;          // no. of events the course has
      for (int j = 0; j < cEvents; ++j)
      {
         event e;
         in >> e.dow;            // when is the course
         in >> e.hour;
         std::string roomcode;
         in >> roomcode;         // where is the course
         if (roomcode != "LAB")
         {
            // a lecture hour; find the room
            UNITS_ITER posU = std::find_if(net[e].units.begin(), net[e].units.end(), find_room(roomcode));
            // room must exist and be unoccupied
            if (posU == net[e].units.end() || !(Subject(posU).code.empty()))
               throw std::runtime_error("board::ReadPermanents");

            posU->subject = *pos;      // place the course in that room
            posU->bPermanent = true;   // this is permanent course
         }
         else
         {
            // a lab hour; just create/append a new unit for lab room and course
            unit u;
            u.subject = *pos;          // set all to lab properties
            u.place.bLab = true;
            u.place.code = "LAB";
            u.place.name = "LAB";
            u.place.size = u.subject.size;
            u.place.owner = u.subject.owner;
            u.bPermanent = true;       // this is permanent course
            net[e].units.push_back(u);
         }
      }
   }

   in.exceptions(oldEx);
}

void board::ReadClashes(std::ifstream &in)
{
   // save current state of exception flags; this function can throw exceptions
   std::ios_base::iostate oldEx = in.exceptions();
   in.exceptions(std::ios_base::failbit | std::ios_base::badbit | std::ios_base::eofbit);

   int cTotalClash;
   in >> cTotalClash;
   for (int i = 0; i < cTotalClash; ++i)
   {
      int cClash;
      in >> cClash;
      std::vector<std::string> clash;
      int j = 0;
      for (; j < cClash; ++j)       // collect all clashes per line
      {
         std::string code;
         in >> code;
         clash.push_back(code);
      }

      for (j = 0; j < cClash; ++j)  
      {
         // insert the clashes into the respective course's clash list
         course tCourse;
         tCourse.code = clash[j];   // only course code is used for comparison
         COURSE_ITER pos = std::find(pts.begin(), pts.end(), tCourse);
         if (pos != pts.end())
         {
            // copy all clashes except itself
            copy_if(clash.begin(), clash.end(), std::back_inserter(pos->clashgrp), no_self_copy(pos->code));
         }
      }
   }

   if (cTotalClash > 0)    
   {
      // remove duplicates in the clash list
      for (COURSE_ITER pos=pts.begin(); pos!=pts.end(); ++pos)
      {
         std::sort(pos->clashgrp.begin(), pos->clashgrp.end());
         pos->clashgrp.erase(std::unique(pos->clashgrp.begin(), pos->clashgrp.end()), pos->clashgrp.end());
      }
   }

   in.exceptions(oldEx);
}

