#ifndef BOARD_H
#define BOARD_H

#include <cstddef>
#include <cstdlib>
#include <map>
#include <fstream>
#include "row.h"

///////////////////////////////////////////////////////////////////////////////
// random - slightly better random number generator
///////////////////////////////////////////////////////////////////////////////

struct random
{
public:
   std::ptrdiff_t operator()(std::ptrdiff_t max=32767)
   {
      double dTmp = static_cast<double>(std::rand()) / static_cast<double>(32767);
      return static_cast<std::ptrdiff_t>(dTmp * max);
   }
};

///////////////////////////////////////////////////////////////////////////////
// event
///////////////////////////////////////////////////////////////////////////////

struct event
{
   int dow;    // day of week (1=Mon, 2=Tue, etc.)
   int hour;   // hour

   event(int iDow=1, int iHour=8) : dow(iDow), hour(iHour) {}

   bool operator<(event const &rhs) const
   {
      return ((dow<rhs.dow) || (dow==rhs.dow && hour<rhs.hour));
   }

   bool operator==(event const &rhs) const 
   {
      return (dow==rhs.dow && hour==rhs.hour);
   }
};

///////////////////////////////////////////////////////////////////////////////
// BOARDFILES
///////////////////////////////////////////////////////////////////////////////

struct BOARDFILES
{
private:
   // disallow copying and assignment (no files sharing)
   BOARDFILES(BOARDFILES const &rhs);
   BOARDFILES &operator=(BOARDFILES const &rhs);

public:
   std::ifstream fEvent;         // event file
   std::ifstream fRoom;          // room list file
   std::ifstream fCourse;        // course list file
   std::ifstream fPermanent;     // permanent course list file
   std::ifstream fClash;         // clash groups file
   
   BOARDFILES() {}
   ~BOARDFILES()
   {
      if (fEvent) fEvent.close();
      if (fRoom) fRoom.close();
      if (fCourse) fCourse.close();
      if (fPermanent) fPermanent.close();
      if (fClash) fClash.close();
   }
};

///////////////////////////////////////////////////////////////////////////////
// board
///////////////////////////////////////////////////////////////////////////////

class board
{
public:
   // Helpers:
   typedef std::map<event, row>::iterator       MAP_ITER;
   typedef std::map<event, row>::const_iterator MAP_C_ITER;
   typedef std::vector<unit>::iterator          UNITS_ITER;
   typedef std::vector<unit>::const_iterator    UNITS_C_ITER;
   typedef std::vector<course>::iterator        COURSE_ITER;
   typedef std::vector<course>::const_iterator  COURSE_C_ITER;
   
   static event const &Event(MAP_ITER &pos);
   static event const &Event(MAP_C_ITER &pos);
   static int const &Dow(MAP_ITER &pos);
   static int const &Dow(MAP_C_ITER &pos);
   static int const &Hour(MAP_ITER &pos);
   static int const &Hour(MAP_C_ITER &pos);
   static row &Row(MAP_ITER &pos);
   static const row &Row(MAP_C_ITER &pos);
   static std::vector<unit> &Units(MAP_ITER &pos);
   static std::vector<unit> const &Units(MAP_C_ITER &pos);
   static course &Subject(UNITS_ITER &pos);
   static course const &Subject(UNITS_C_ITER &pos);
   static room &Place(UNITS_ITER &pos);
   static room const &Place(UNITS_C_ITER &pos);

private:
   struct ROOMIDX_
   {
      int first, last;
      ROOMIDX_(int frst=-1, int lst=-1) : first(frst), last(lst) {}
      bool operator==(ROOMIDX_ const &rhs) const
      {
         return (first==rhs.first && last==rhs.last);
      }
   };

   struct OCCURRED_
   {
      int numLec, numLab;
      OCCURRED_(int lec=0, int lab=0) : numLec(lec), numLab(lab) {}
   };

   std::map<int, ROOMIDX_> roomidx_;
   int rowSize_, nIter_, nLimit_;

   bool Count(OCCURRED_ &oc, event const &e, course const &c);
   void ClearAllCourses(bool bLec=true, bool bLab=true);
   bool IsSizeOk() const;
   std::pair<MAP_ITER, UNITS_ITER> const RandomCell(course const &c, bool bLab=false);
   void DelLabUnits(int dow, course c);
   bool AddLabUnits(MAP_ITER pos[3], course const &c);
   bool FitLab(course const &c);
   bool FitLecture(course const &c);
   void ReadPermanents(std::ifstream &in);
   void ReadClashes(std::ifstream &in);
   void ReadCourses(std::ifstream &in);
   void PreFit();
   void PostFit();

public:
   std::vector<course> pts;
   std::map<event, row> net;
   int fitOwner;

   board();
   board(board const &rhs);
   board &operator=(board const &rhs);

   int BoardSize() const;
   std::vector<std::pair<event, unit> > const CourseSchedule(course const &c) const;

   bool Fit(int cAttempts=1);
   bool IsValid(bool bIgnore=false);

   void ReadBoard(BOARDFILES &bf);
};


///////////////////////////////////////////////////////////////////////////////
// board definitions
///////////////////////////////////////////////////////////////////////////////

inline board::board()
   : rowSize_(-1), nIter_(0), nLimit_(1000), fitOwner(-1)
{}

inline board::board(board const &rhs) 
   : rowSize_(rhs.rowSize_), nIter_(rhs.nIter_), nLimit_(rhs.nLimit_), roomidx_(rhs.roomidx_),
     pts(rhs.pts), net(rhs.net), fitOwner(rhs.fitOwner)
{}

inline board &board::operator=(board const &rhs)
{
   rowSize_ = rhs.rowSize_;
   nIter_ = rhs.nIter_;
   nLimit_ = rhs.nLimit_;
   roomidx_ = rhs.roomidx_;
   pts = rhs.pts;
   net = rhs.net;
   fitOwner = rhs.fitOwner;
   return *this;
}

inline event const &board::Event(MAP_ITER &pos)
{
   return pos->first;
}

inline event const &board::Event(MAP_C_ITER &pos)
{
   return pos->first;
}

inline int const &board::Dow(MAP_ITER &pos)
{
   return pos->first.dow;
}

inline int const &board::Dow(MAP_C_ITER &pos)
{
   return pos->first.dow;
}

inline int const &board::Hour(MAP_ITER &pos)
{
   return pos->first.hour;
}

inline int const &board::Hour(MAP_C_ITER &pos)
{
   return pos->first.hour;
}

inline row &board::Row(MAP_ITER &pos)
{
   return pos->second;
}

inline row const &board::Row(MAP_C_ITER &pos)
{
   return pos->second;
}

inline std::vector<unit> &board::Units(MAP_ITER &pos)
{
   return pos->second.units;
}

inline std::vector<unit> const &board::Units(MAP_C_ITER &pos)
{
   return pos->second.units;
}

inline course &board::Subject(UNITS_ITER &pos)
{
   return pos->subject;
}

inline course const &board::Subject(UNITS_C_ITER &pos)
{
   return pos->subject;
}

inline room &board::Place(UNITS_ITER &pos)
{
   return pos->place;
}

inline room const &board::Place(UNITS_C_ITER &pos)
{
   return pos->place;
}

#endif
