#include <algorithm>
#include "row.h"

namespace
{
   struct clear_courses
   {
      course reset;        // blank course
      bool blec, blab;
      clear_courses(bool b0, bool b1) : blec(b0), blab(b1) {}

      void operator()(unit &u) const
      {
         if (!u.bPermanent && ((u.place.bLab && blab) || (!u.place.bLab && blec)))
            u.subject = reset;
      }
   };

   struct get_clashgroups
   {
      std::vector<std::string> *pclash;
      bool bAdd;

      get_clashgroups(std::vector<std::string> &d, bool badd) : pclash(&d), bAdd(badd) {}

      void operator()(unit const &u)
      {
         if (!u.subject.code.empty())
         {
            if (bAdd) pclash->push_back(u.subject.code);
            std::copy(u.subject.clashgrp.begin(), u.subject.clashgrp.end(), std::back_inserter(*pclash));
         }
      }
   };

   struct find_clash
   {
      std::vector<std::string> *pclash;

      find_clash(std::vector<std::string> &d) : pclash(&d) {}

      bool operator()(unit const &u) const
      {
         return (std::find(pclash->begin(), pclash->end(), u.subject.code) != pclash->end());
      }

   };

   struct clash_permanent
   {
      std::vector<unit>::const_iterator posI;
      bool bValid;

      clash_permanent(std::vector<unit>::const_iterator p) : posI(p), bValid(true) {}

      void operator()(unit const &u)
      {
         if (bValid && !(*posI == u))
         {
            if (std::find(u.subject.clashgrp.begin(), u.subject.clashgrp.end(),
                          posI->subject.code) != u.subject.clashgrp.end())
               bValid = (u.bPermanent == posI->bPermanent == true);
         }
      }
   };
}

// Clears courses from all units:
//    bLec = true means clear all lecture credits
//    bLab = true means clear all lab credits
void row::ClearCourses(bool bLec, bool bLab)
{
   std::for_each(units.begin(), units.end(), clear_courses(bLec, bLab));
}

// Returns a list of clash groups from all units:
//    bAdd = true means add the course codes into the list as well
std::vector<std::string> const row::GroupClashes(bool bAdd) const
{
   std::vector<std::string> clash;
   std::for_each(units.begin(), units.end(), get_clashgroups(clash, bAdd));
   return clash;
}

// Returns true if all units are valid (no clashes)
//    bIgnore - set to true to fully check any course clashes; else set to
//              false to ignore clashes between permanent courses
bool row::IsValid(bool bIgnore) const
{
   std::vector<std::string> clash = GroupClashes(false);   // get list of clashes
   // any clash?
   bool bValid =(std::find_if(units.begin(), units.end(), find_clash(clash)) == units.end());
   if (!bValid && bIgnore)
   {
      // clashes found, but determine if clashes are between 2 or more permanent courses
      bool bOk = true;
      for (std::vector<unit>::const_iterator posI=units.begin(); bOk && posI != units.end(); ++posI)
      {
         clash_permanent cp(posI);
         std::for_each(units.begin(), units.end(), cp);
         bOk = cp.bValid;
      }
      bValid = bOk;
   }

   return bValid;
}
